@@ -6,7 +6,7 @@
 //
 /*----------------------------------------------------------------------------*/
 /*******************************************************************************
-Copyright (C) 2022 Steve Jeong <jkhpro1003@gmail.com>
+Copyright (C) 2022 Steve Jeong <how2soft@gmail.com>
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
@@ -21,6 +21,7 @@ GNU General Public License for more details.
 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
+#include <dirent.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <stdint.h>
@@ -31,6 +32,7 @@ along with this program.  If not, see <http://www.gnu.org/licenses/>.
 #include <sys/ioctl.h>
 #include <asm/ioctl.h>
 #include <sys/mman.h>
+#include <sys/stat.h>
 /*----------------------------------------------------------------------------*/
 #include "softPwm.h"
 #include "softTone.h"
@@ -103,6 +105,68 @@ static const int phyToGpio[64] = {
 	-1, -1, -1, -1, -1, -1, -1, -1,	// 49...56
 	-1, -1, -1, -1, -1, -1, -1	// 57...63
 };
+
+static const char *pinToPwm[64] = {
+	// wiringPi number to pwm group number
+		"None", "None",//  0 |  1 : GPIO0_C0, GPIO3_D0
+		"None", "fe6f0010",	   //  2 |  3 : GPIO0_C1, GPIO3_B2(PWM9)
+		"None", "None",		   //  4 |  5 : GPIO3_C6, GPIO3_C7
+		"None", "fdd70020",	   //  6 |  7 : GPIO3_D1, GPIO0_B6(PWM2)
+		"None", "None",        //  8 |  9 : GPIO3_B6, GPIO3_B5
+		"None", "None",        // 10 | 11 : GPIO2_D2, GPIO3_D2
+		"None", "None",        // 12 | 13 : GPIO2_D1, GPIO2_D0
+		"None", "None",        // 14 | 15 : GPIO2_D3, GPIO3_D6
+		"None", "None",		   // 16 | 17 : GPIO3_D7
+		"None", "None",		   // 18 | 19 :
+		"None", "None",		   // 20 | 21 : , GPIO4_C1
+		"None", "fdd70010",	   // 22 | 23 : GPIO4_B6, GPIO0_B5(PWM1)
+		"None", "None",	   // 24 | 25 : GPIO3_D5,
+		"None", "None",        // 26 | 27 : GPIO3_D3, GPIO3_D4
+		"None", "None",        // 28 | 29 :
+		"None", "None",        // 30 | 31 : GPIO0_B4, GPIO0_B3
+	// Padding:
+	"None","None","None","None","None","None","None","None","None","None","None","None","None","None","None","None", // 32...47
+	"None","None","None","None","None","None","None","None","None","None","None","None","None","None","None","None"  // 48...63
+};
+
+static const int pinToPwmNum[64] = {
+	// wiringPi number to pwm pin number
+	 -1, -1,	//  0 |  1 : GPIO0_C0, GPIO3_D0
+	 -1,  2,	//  2 |  3 : GPIO0_C1, GPIO3_B2(PWM9)
+	 -1, -1,	//  4 |  5 : GPIO3_C6, GPIO3_C7
+	 -1,  1,	//  6 |  7 : GPIO3_D1, GPIO0_B6(PWM2)
+	 -1, -1,	//  8 |  9 : GPIO3_B6, GPIO3_B5
+	 -1, -1,	// 10 | 11 : GPIO2_D2, GPIO3_D2
+	 -1, -1,	// 12 | 13 : GPIO2_D1, GPIO2_D0
+	 -1, -1,	// 14 | 15 : GPIO2_D3, GPIO3_D6
+	 -1, -1,	// 16 | 17 : GPIO3_D7
+	 -1, -1,	// 18 | 19 :
+	 -1, -1,	// 20 | 21 : , GPIO4_C1
+	 -1,  0,	// 22 | 23 : GPIO4_B6, GPIO0_B5(PWM1)
+	 -1, -1,	// 24 | 25 : GPIO3_D5,
+	 -1, -1,	// 26 | 27 : GPIO3_D3, GPIO3_D4
+	 -1, -1,	// 28 | 29 :
+	 -1, -1,	// 30 | 31 : GPIO0_B4, GPIO0_B3
+	// Padding:
+	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,	// 32...47
+	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1	// 48...63
+};
+
+static char pwmPinPath[10][(BLOCK_SIZE)] = {
+	"","",
+	"",
+	// Padding:
+	"None","None","None",
+	"None","None","None","None"
+};
+
+static char setupedPwmPinPath[10][BLOCK_SIZE] = {
+	"None","None",
+	"None","None",
+	"None","None",
+	"None","None",
+	"None","None"
+};
 /*----------------------------------------------------------------------------*/
 //
 // Global variable define
@@ -123,13 +187,30 @@ static volatile uint32_t *cru[2];
 /* wiringPi Global library */
 static struct libodroid	*lib = NULL;
 
+/* pwm sysnode */
+static DIR *pwm;
+static struct dirent *pwmchip;
+/* pwm params */
+static char sysPwmPath[(BLOCK_SIZE / 4)];
+static char pwmExport[(BLOCK_SIZE / 16)];
+static char pwmUnexport[(BLOCK_SIZE / 16)];
+static char pwmPeriod[(BLOCK_SIZE / 16)];
+static char pwmDuty[(BLOCK_SIZE / 16)];
+static unsigned int pwmClock;
+static unsigned int pwmRange;
 /*----------------------------------------------------------------------------*/
 // Function prototype define
 /*----------------------------------------------------------------------------*/
 static int	gpioToShiftRegBy32	(int pin);
 static int	gpioToShiftRegBy16	(int pin);
 static void	setClkState	(int pin, int state);
-static void	setIomuxMode 	(int pin, int mode);
+static int	setIomuxMode 	(int pin, int mode);
+/*----------------------------------------------------------------------------*/
+// Function of pwm define
+/*----------------------------------------------------------------------------*/
+static int	pinToSysPwmPath	(int pin);
+static int	pwmSetup (int pin);
+static int	pwmRelease (int pin);
 /*----------------------------------------------------------------------------*/
 // wiringPi core function
 /*----------------------------------------------------------------------------*/
@@ -142,9 +223,12 @@ static int		_getPUPD		(int pin);
 static int		_pullUpDnControl	(int pin, int pud);
 static int		_digitalRead		(int pin);
 static int		_digitalWrite		(int pin, int value);
+static int		_pwmWrite		(int pin, int value);
 static int		_analogRead		(int pin);
 static int		_digitalWriteByte	(const unsigned int value);
 static unsigned int	_digitalReadByte	(void);
+static void		_pwmSetRange	(unsigned int range);
+static void		_pwmSetClock	(int divisor);
 
 #if !defined(DEVMEM)
 /*----------------------------------------------------------------------------*/
@@ -182,6 +266,114 @@ static int gpioToShiftRegBy16 (int pin)
 	return pin % 16;
 }
 /*----------------------------------------------------------------------------*/
+//
+// config pwm sys path. "/sys/class/pwm/pwmchip?"
+//
+/*----------------------------------------------------------------------------*/
+static int pinToSysPwmPath (int pin)
+{
+	const char *pwmGroup;
+	char pwmLinkSrc[(BLOCK_SIZE / 8)];
+	char pwmPath[(BLOCK_SIZE / 8)];
+	int sz_link;
+
+	memset(pwmLinkSrc, 0, sizeof(pwmLinkSrc));
+	memset(pwmPath, 0, sizeof(pwmPath));
+
+	pwmGroup = pinToPwm[pin];
+	pwm = opendir("/sys/class/pwm");
+	if (pwm == NULL) {
+		printf("need to set device: pwm\n");
+		return -1;
+	}
+
+	while (1) {
+		pwmchip = readdir(pwm);
+
+		if (pwmchip == NULL) {
+			break;
+		}
+
+		if (strlen(pwmchip->d_name) <= 2)
+			continue;
+
+		sprintf(pwmPath, "%s/%s", "/sys/class/pwm", pwmchip->d_name);
+		sz_link = readlink(pwmPath, pwmLinkSrc, sizeof(pwmLinkSrc));
+		if (sz_link < 0) {
+			perror("Read symbolic link fail");
+			return sz_link;
+		}
+
+		if (strstr(pwmLinkSrc, pwmGroup) != NULL) {
+			strncpy(sysPwmPath, pwmPath, (sizeof(sysPwmPath) - 1));
+			break;
+		}
+	}
+	closedir(pwm);
+
+	return 0;
+}
+
+static int pwmSetup (int pin) {
+	char cmd[(BLOCK_SIZE * 2)];
+	int pwmPin, ret;
+
+	memset(cmd, 0, sizeof(cmd));
+	memset(pwmExport, 0, sizeof(pwmExport));
+
+	if ((ret = pinToSysPwmPath(pin)) < 0) {
+		perror("set pwm dtb overlays");
+		return ret;
+	}
+
+	if (strstr(sysPwmPath, "pwmchip") == NULL) {
+		printf("config pwm dtb overlays\n");
+		return -1;
+	}
+
+	pwmPin = pinToPwmNum[pin];
+	if (pwmPin == 2)
+		pwmClock = M1_PWM_INTERNAL_CLK;
+	else
+		pwmClock = (M1_PWM_INTERNAL_CLK / 2);
+	sprintf(pwmExport, "%d", 0);
+	sprintf(pwmPinPath[pwmPin], "%s/pwm%d", sysPwmPath, 0);
+	strncpy(setupedPwmPinPath[pwmPin], pwmPinPath[pwmPin], (BLOCK_SIZE - 1));
+#ifdef ANDROID
+	sprintf(cmd, "su -s sh -c %s %s", PWM_ACCESS_SCRIPT, pwmPinPath[pwmPin]);
+#else
+	sprintf(cmd, "sudo sh %s %s", PWM_ACCESS_SCRIPT, pwmPinPath[pwmPin]);
+#endif
+	inputToSysNode(sysPwmPath, "export", pwmExport);
+	system(cmd);
+	printf("PWM/pin%d: Don't change to gpio mode with overlay registered.\n", pin);
+
+	return 0;
+}
+
+static int pwmRelease (int pin) {
+	int pwmPin, ret;
+
+	if ((ret = pinToSysPwmPath(pin)) < 0) {
+		return ret;
+	}
+
+	if (strstr(sysPwmPath, "pwmchip") == NULL) {
+		return -1;
+	}
+
+	pwmPin = pinToPwmNum[pin];
+	sprintf(pwmUnexport, "%d", (pwmPin % 2));
+	sprintf(pwmPinPath[pwmPin], "%s/pwm%d", sysPwmPath, (pwmPin % 2));
+	if ((pwm = opendir(pwmPinPath[pwmPin])) != NULL) {
+		inputToSysNode(pwmPinPath[pwmPin], "enable", "0");
+		inputToSysNode(sysPwmPath, "unexport", pwmUnexport);
+		closedir(pwm);
+	}
+
+	return 0;
+}
+/*----------------------------------------------------------------------------*/
 static int _getModeToGpio (int mode, int pin)
 {
 	if (pin > 255)
@@ -258,11 +450,16 @@ __attribute__ ((unused))static void setClkState (int pin, int state)
 // set IOMUX mode
 //
 /*----------------------------------------------------------------------------*/
-__attribute__ ((unused))static void setIomuxMode (int pin, int mode)
+__attribute__ ((unused))static int setIomuxMode (int pin, int mode)
 {
 	uint32_t offset, target;
 	uint8_t	bank, group, bitNum, bitInByte;
 
+	if (lib->mode == MODE_GPIO_SYS)
+		return -1;
+	if ((pin = _getModeToGpio(lib->mode, pin)) < 0)
+		return -1;
+
 	bank = pin / GPIO_SIZE;
 	bitNum = pin - (bank * GPIO_SIZE);
 	group = bitNum / 8; // A or B or C or D
@@ -273,7 +470,7 @@ __attribute__ ((unused))static void setIomuxMode (int pin, int mode)
 
 	// Common IOMUX Funtion 1 : GPIO (3'h0)
 	switch (mode) {
-	case M1_FUNC_GPIO://Common IOMUX Function 1_GPIO (3'h0)
+	case M1_FUNC_GPIO: // Common IOMUX Function 1_GPIO (3'h0)
 		if (bank == 0) {
 			offset += M1_PMU_GRF_IOMUX_OFFSET;
 			target = *(grf[0] + (offset >> 2));
@@ -289,18 +486,36 @@ __attribute__ ((unused))static void setIomuxMode (int pin, int mode)
 			*(grf[1] + (offset >> 2)) = target;
 		}
 		break;
+	case M1_FUNC_PWM:
+		if (bank == 0) {
+			offset += M1_PMU_GRF_IOMUX_OFFSET;
+			target = *(grf[0] + (offset >> 2));
+			target |= (0x7 << ((bitInByte % 4) * 4 + 16));
+			target |= (0x4 << ((bitInByte % 4) * 4)); // gpio0 b5/b6: 3'h100
+			*(grf[0] + (offset >> 2)) = target;
+		}
+		else {
+			offset += M1_SYS_GRF_IOMUX_OFFSET;
+			target = *(grf[1] + (offset >> 2));
+			target |= (0x7 << ((bitInByte % 4) * 4 + 16));
+			target |= (0x5 << ((bitInByte % 4) * 4)); // gpio3 b2: 3'h101
+			*(grf[1] + (offset >> 2)) = target;
+		}
+		break;
 	default:
 		break;
 	}
+
+	return 0;
 }
 /*----------------------------------------------------------------------------*/
 __attribute__ ((unused))static int _pinMode (int pin, int mode)
 {
 	uint32_t offset, target;
 	uint8_t bank, bitNum;
-	int oriPin;
+	int origPin;
 
-	oriPin = pin;
+	origPin = pin;
 
 	if (lib->mode == MODE_GPIO_SYS)
 		return -1;
@@ -312,6 +527,7 @@ __attribute__ ((unused))static int _pinMode (int pin, int mode)
 	bitNum = pin - (bank * GPIO_SIZE);
 	offset = bitNum / 16 == 0 ? M1_GPIO_DIR_OFFSET : M1_GPIO_DIR_OFFSET + 0x4;
 
+	pwmRelease (origPin);
 	softPwmStop(pin);
 	softToneStop(pin);
 
@@ -320,29 +536,36 @@ __attribute__ ((unused))static int _pinMode (int pin, int mode)
 
 	switch (mode) {
 	case INPUT:
+		setIomuxMode(origPin, M1_FUNC_GPIO);
 		target &= ~(1 << gpioToShiftRegBy16(pin));
 		*(gpio[bank] + (offset >> 2)) = target;
-		_pullUpDnControl(oriPin, PUD_OFF);
+		_pullUpDnControl(origPin, PUD_OFF);
 		break;
 	case OUTPUT:
+		setIomuxMode(origPin, M1_FUNC_GPIO);
 		target |= (1 << gpioToShiftRegBy16(pin));
 		*(gpio[bank] + (offset >> 2)) = target;
 		break;
 	case INPUT_PULLUP:
+		setIomuxMode(origPin, M1_FUNC_GPIO);
 		target &= ~(1 << gpioToShiftRegBy16(pin));
 		*(gpio[bank] + (offset >> 2)) = target;
-		_pullUpDnControl(oriPin, PUD_UP);
+		_pullUpDnControl(origPin, PUD_UP);
 		break;
 	case INPUT_PULLDOWN:
+		setIomuxMode(origPin, M1_FUNC_GPIO);
 		target &= ~(1 << gpioToShiftRegBy16(pin));
 		*(gpio[bank] + (offset >> 2)) = target;
-		_pullUpDnControl(oriPin, PUD_DOWN);
+		_pullUpDnControl(origPin, PUD_DOWN);
 		break;
 	case SOFT_PWM_OUTPUT:
-		softPwmCreate(oriPin, 0, 100);
+		softPwmCreate(origPin, 0, 100);
 		break;
 	case SOFT_TONE_OUTPUT:
-		softToneCreate(oriPin);
+		softToneCreate(origPin);
+		break;
+	case	PWM_OUTPUT:
+		pwmSetup(origPin);
 		break;
 	default:
 		msg(MSG_WARN, "%s : Unknown Mode %d\n", __func__, mode);
@@ -921,6 +1144,29 @@ __attribute__ ((unused))static int _digitalWrite_gpiod (int pin, int value)
 }
 #endif
 
+__attribute__ ((unused))static int _pwmWrite (int pin, int value)
+{
+	unsigned int duty;
+	int pwmPin;
+
+	memset(pwmDuty, 0, sizeof(pwmDuty));
+
+	if (lib->mode == MODE_GPIO_SYS)
+		return -1;
+
+	if (((unsigned int)value > pwmRange) || (pwmRange <= 0)) {
+		printf("warn : pwm range value is greater than or equal pwmWrite's\n");
+		return -1;
+	}
+
+	pwmPin = pinToPwmNum[pin];
+	duty = ((value * 100) / pwmRange);
+	sprintf(pwmDuty, "%d", ((atoi(pwmPeriod) * duty) / 100));
+
+	inputToSysNode(pwmPinPath[pwmPin], "duty_cycle", pwmDuty);
+
+	return 0;
+}
 /*----------------------------------------------------------------------------*/
 static int _analogRead (int pin)
 {
@@ -1175,6 +1421,63 @@ __attribute__ ((unused))static unsigned int _digitalReadByte_gpiod (void)
 }
 #endif
 
+/*----------------------------------------------------------------------------*/
+// PWM signal ___-----------___________---------------_______-----_
+//               <--value-->           <----value---->
+//               <-------range--------><-------range-------->
+// PWM frequency == (PWM clock) / range
+/*----------------------------------------------------------------------------*/
+static void _pwmSetRange (unsigned int range)
+{
+	unsigned int freq, period;
+
+	memset(pwmPeriod, 0, sizeof(pwmPeriod));
+
+	if (lib->mode == MODE_GPIO_SYS)
+		return;
+
+	if (pwmClock < 2) {
+		printf("error : pwm1 or pwm2 freq: %dMHz / (pwmSetClock's value) >= 2\n",
+				(M1_PWM_INTERNAL_CLK / 2000000));
+		printf("error : pwm9 freq: %dMHz / (pwmSetClock's value) >= 2\n",
+				(M1_PWM_INTERNAL_CLK / 1000000));
+		return;
+	}
+
+	pwmRange = range;
+	if ((pwmRange < 1) || (pwmRange >= pwmClock)) {
+		printf("error : invalid value. ( < pwm freq)\n");
+		return;
+	}
+
+	freq = (pwmClock / pwmRange);
+	period = (1000000000 / freq); // period: s to ns.
+	sprintf(pwmPeriod, "%d", period);
+
+	for (int i = 0; i < 10; i++) {
+		if (strstr(setupedPwmPinPath[i], "None") != NULL)
+			continue;
+		inputToSysNode(setupedPwmPinPath[i], "period", pwmPeriod);
+		inputToSysNode(setupedPwmPinPath[i], "polarity", "normal");
+		inputToSysNode(setupedPwmPinPath[i], "enable", "1");
+	}
+}
+
+/*----------------------------------------------------------------------------*/
+// Internal clock: PWM0: 12MHz PWM2: 24MHz
+// PWM clock == (Internal clock) / divisor
+// PWM frequency == (PWM clock) / range
+/*----------------------------------------------------------------------------*/
+static void _pwmSetClock (int divisor)
+{
+	if (pwmClock > 0)
+		pwmClock = (pwmClock / divisor);
+	else {
+		printf("error : pwm mode error\n");
+		return;
+	}
+}
+
 /*----------------------------------------------------------------------------*/
 static void init_gpio_mmap (void)
 {
@@ -1219,15 +1522,16 @@ static void init_gpio_mmap (void)
 		} else {
 			cru[0] = (uint32_t *) mapped_cru[0];
 			cru[1] = (uint32_t *) mapped_cru[1];
-			}
+		}
+
 		if ((mapped_grf[0] == MAP_FAILED) || (mapped_grf[1] == MAP_FAILED)) {
 			msg (MSG_ERR,"wiringPiSetup: mmap (GRF) failed: %s\n",strerror (errno));
 		} else {
 			grf[0] = (uint32_t *) mapped_grf[0];
 			grf[1] = (uint32_t *) mapped_grf[1];
-			}
+		}
 
-		if (	(mapped_gpio[0] == MAP_FAILED) ||
+		if ((mapped_gpio[0] == MAP_FAILED) ||
 			(mapped_gpio[1] == MAP_FAILED) ||
 			(mapped_gpio[2] == MAP_FAILED) ||
 			(mapped_gpio[3] == MAP_FAILED) ||
@@ -1275,6 +1579,9 @@ void init_odroidm1 (struct libodroid *libwiring)
 	libwiring->analogRead		= _analogRead;
 	libwiring->digitalWriteByte	= _digitalWriteByte;
 	libwiring->digitalReadByte	= _digitalReadByte;
+	libwiring->pwmWrite			= _pwmWrite;
+	libwiring->pwmSetRange		= _pwmSetRange;
+	libwiring->pwmSetClock		= _pwmSetClock;
 #else
 	/* wiringPi-libgpiod Core function initialize */
 	libwiring->getModeToGpio	= _getModeToGpio;
