@@ -5,6 +5,7 @@
 //
 //
 /*----------------------------------------------------------------------------*/
+#include <dirent.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <stdint.h>
@@ -15,6 +16,7 @@
 #include <sys/ioctl.h>
 #include <asm/ioctl.h>
 #include <sys/mman.h>
+#include <sys/stat.h>
 
 /*----------------------------------------------------------------------------*/
 #include "softPwm.h"
@@ -79,80 +81,66 @@ static const int phyToGpio_rev1[64] = {
 	-1, -1, -1, -1, -1, -1, -1	// 57...63
 };
 
-static int16_t _gpioTophysPin [] = {
-			// (native gpio number - N2_GPIOA_PIN_START) to physical header pin number
-	 -1,		// 			0 + N2_GPIOA_PIN_START(460)
-	 -1,  -1,	// 			1	| 2
-	 -1,  26,	// 			3	| 4	GPIOA.4(SPI_CE1)
-	 -1,  -1,	// 			5	| 6
-	 -1,  -1,	// 			7	| 8
-	 -1,  -1,	// 			9	| 10
-	 -1,  32,	// 			11	| 12	GPIOA.12
-	  7,  27,	// GPIOA.13		13	| 14	GPIOA.14(I2C-3_SDA)
-	 28,  16,	// GPIOA.15(I2C-3_SCL)	15	| 16	GPIOX.0
-	 18,  22,	// GPIOX.1		17	| 18	GPIOX.2
-	 11,  13,	// GPIOX.3		19	| 20	GPIOX.4
-	 33,  35,	// GPIOX.5(PWM_C)	21	| 22	GPIOX.6(PWM_D)
-	 15,  19,	// GPIOX.7(PWM_F)	23	| 24	GPIOX.8(SPI_MOSI)
-	 21,  24,	// GPIOX.9(SPI_MISO)	25	| 26	GPIOX.10(SPI_CE0)
-	 23,   8,	// GPIOX.11(SPI_SCLK)	27	| 28	GPIOX.12(UART_TX_B)
-	 10,  29,	// GPIOX.13(UART_RX_B)	29	| 30	GPIOX.14
-	 31,  12,	// GPIOX.15		31	| 32	GPIOX.16(PWM_E)
-	  3,   5,	// GPIOX.17(I2C-2_SDA)	33	| 34	GPIOX.18(I2C-2_SCL)
-	 36,  -1,	// GPIOX.19		35	| 36
-	 -1,  -1,	// 			37	| 38
-	 -1,  -1,	// 			39	| 40
-	// Not used
-	-1, -1, -1, -1, -1, -1, -1, -1,	// 41...48
-	-1, -1, -1, -1, -1, -1, -1, -1,	// 49...56
-	-1, -1, -1, -1, -1, -1, -1	// 57...63
+static const char *pinToPwm[64] = {
+	// wiringPi number to pwm group number
+		"None", "ffd19000", // GPIOX.3			0	| 1	GPIOX.16(PWM_E)
+		"None", "ffd19000",		// GPIOX.4			2	| 3	GPIOX.7(PWM_F)
+		"None", "None",			// GPIOX.0			4	| 5	GPIOX.1
+		"None", "None",		// GPIOX.2			6	| 7	GPIOA.13
+		"None", "None",			// GPIOX.17(I2C-2_SDA)		8	| 9	GPIOX.18(I2C-2_SCL)
+		"None", "None",			// GPIOX.10			10	| 11	GPIOA.4
+		"None", "None",			// GPIOX.8			12	| 13	GPIOX.9
+		"None", "None",			// GPIOX.11			14	| 15	GPIOX.12
+		"None", "None",			// GPIOX.13			16	| 17
+		"None", "None",			// 				18	| 19
+		"None", "None",			// 				20	| 21	GPIOX.14
+		"None", "ffd1a000",     // GPIOX.15			22	| 23	GPIOX.5(PWM_C)
+		"ffd1a000", "None",     // GPIOX.6(PWM_D)		24	| 25	ADC.AIN3
+		"None", "None",         // GPIOA.12			26	| 27	GPIOX.19
+		"None", "None",         // REF1.8V OUT			28	| 29	ADC.AIN2
+		"None", "None",         // GPIOA.14(I2C-3_SDA)		30	| 31	GPIOA.15(I2C-3_SCL)
+	// Padding:
+	"None","None","None","None","None","None","None","None","None","None","None","None","None","None","None","None", // 32...47
+	"None","None","None","None","None","None","None","None","None","None","None","None","None","None","None","None"  // 48...63
 };
 
-static int8_t _gpioToPwmPin [] = {
-			// (native gpio number - N2_GPIOA_PIN_START) to PWM pin number
-	 -1,		// 			0 + N2_GPIOA_PIN_START(460)
-	 -1,  -1,	// 			1	| 2
-	 -1,  -1,	// 			3	| 4	GPIOA.4(SPI_CE1)
-	 -1,  -1,	// 			5	| 6
-	 -1,  -1,	// 			7	| 8
-	 -1,  -1,	// 			9	| 10
-	 -1,  -1,	// 			11	| 12	GPIOA.12
-	 -1,  -1,	// GPIOA.13		13	| 14	GPIOA.14(I2C-3_SDA)
-	 -1,  -1,	// GPIOA.15(I2C-3_SCL)	15	| 16	GPIOX.0
-	 -1,  -1,	// GPIOX.1		17	| 18	GPIOX.2
-	 -1,  -1,	// GPIOX.3		19	| 20	GPIOX.4
-	  2,   3,	// GPIOX.5(PWM_C)	21	| 22	GPIOX.6(PWM_D)
-	  5,  -1,	// GPIOX.7(PWM_F)	23	| 24	GPIOX.8(SPI_MOSI)
-	 -1,  -1,	// GPIOX.9(SPI_MISO)	25	| 26	GPIOX.10(SPI_CE0)
-	 -1,  -1,	// GPIOX.11(SPI_SCLK)	27	| 28	GPIOX.12(UART_TX_B)
-	 -1,  -1,	// GPIOX.13(UART_RX_B)	29	| 30	GPIOX.14
-	 -1,   4,	// GPIOX.15		31	| 32	GPIOX.16(PWM_E)
-	 -1,  -1,	// GPIOX.17(I2C-2_SDA)	33	| 34	GPIOX.18(I2C-2_SCL)
-	 -1,  -1,	// GPIOX.19		35	| 36
-	 -1,  -1,	// 			37	| 38
-	 -1,  -1,	// 			39	| 40
-	// Not used
-	-1, -1, -1, -1, -1, -1, -1, -1,	// 41...48
-	-1, -1, -1, -1, -1, -1, -1, -1,	// 49...56
-	-1, -1, -1, -1, -1, -1, -1	// 57...63
+static const int pinToPwmNum[64] = {
+	// wiringPi number to pwm pin number
+	 -1,  2,	// GPIOX.3			0	| 1	GPIOX.16(PWM_E)
+	 -1,  3,	// GPIOX.4			2	| 3	GPIOX.7(PWM_F)
+	 -1, -1,	// GPIOX.0			4	| 5	GPIOX.1
+	 -1, -1,	// GPIOX.2			6	| 7	GPIOA.13
+	 -1, -1,	// GPIOX.17(I2C-2_SDA)		8	| 9	GPIOX.18(I2C-2_SCL)
+	 -1, -1,	// GPIOX.10			10	| 11	GPIOA.4
+	 -1, -1,	// GPIOX.8			12	| 13	GPIOX.9
+	 -1, -1,	// GPIOX.11			14	| 15	GPIOX.12
+	 -1, -1,	// GPIOX.13			16	| 17
+	 -1, -1,	// 				18	| 19
+	 -1, -1,	// 				20	| 21	GPIOX.14
+	 -1,  0,	// GPIOX.15			22	| 23	GPIOX.5(PWM_C)
+	  1, -1,	// GPIOX.6(PWM_D)		24	| 25	ADC.AIN3
+	 -1, -1,	// GPIOA.12			26	| 27	GPIOX.19
+	 -1, -1,	// REF1.8V OUT			28	| 29	ADC.AIN2
+	 -1, -1,	// GPIOA.14(I2C-3_SDA)		30	| 31	GPIOA.15(I2C-3_SCL)
+	// Padding:
+	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,	// 32...47
+	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1	// 48...63
 };
 
-static uint16_t pwmPinToALT [] = {
-	0, 0,	// A, B
-	4, 4,	// C 481 GPIOX.5 , D 482 GPIOX.6
-	1, 1	// E 492 GPIOX.16, F 483 GPIOX.7
+static char pwmPinPath[10][(BLOCK_SIZE)] = {
+	"","",
+	"","",
+	// Padding:
+	"None","None",
+	"None","None","None","None"
 };
 
-static uint16_t pwmPinToRange [] = {
-	0, 0,	// A, B
-	0, 0,	// C 481 GPIOX.5 , D 482 GPIOX.6
-	0, 0	// E 492 GPIOX.16, F 483 GPIOX.7
-};
-
-static uint16_t pwmPinToDutyOffset [] = {
-	N2_PWM_0_DUTY_CYCLE_OFFSET, N2_PWM_1_DUTY_CYCLE_OFFSET,	// A, B
-	N2_PWM_0_DUTY_CYCLE_OFFSET, N2_PWM_1_DUTY_CYCLE_OFFSET,	// C 481 GPIOX.5 , D 482 GPIOX.6
-	N2_PWM_0_DUTY_CYCLE_OFFSET, N2_PWM_1_DUTY_CYCLE_OFFSET	// E 492 GPIOX.16, F 483 GPIOX.7
+static char setupedPwmPinPath[10][BLOCK_SIZE] = {
+	"None","None",
+	"None","None",
+	"None","None",
+	"None","None",
+	"None","None"
 };
 
 /*----------------------------------------------------------------------------*/
@@ -169,11 +157,21 @@ static int adcFds[2];
 
 /* GPIO mmap control */
 static volatile uint32_t *gpio;
-static volatile uint32_t *pwm[3];
 
 /* wiringPi Global library */
 static struct libodroid	*lib = NULL;
 
+/* pwm sysnode */
+static DIR *pwm;
+static struct dirent *pwmchip;
+/* pwm params */
+static char sysPwmPath[(BLOCK_SIZE / 4)];
+static char pwmExport[(BLOCK_SIZE / 16)];
+static char pwmUnexport[(BLOCK_SIZE / 16)];
+static char pwmPeriod[(BLOCK_SIZE / 16)];
+static char pwmDuty[(BLOCK_SIZE / 16)];
+static unsigned int pwmClock;
+static unsigned int pwmRange;
 /*----------------------------------------------------------------------------*/
 // Function prototype define
 /*----------------------------------------------------------------------------*/
@@ -185,9 +183,12 @@ static int	gpioToShiftReg	(int pin);
 static int	gpioToGPFSELReg	(int pin);
 static int	gpioToDSReg	(int pin);
 static int	gpioToMuxReg	(int pin);
-static int	gpioToPwmPin	(int pin);
-static int	gpioTophysPin	(int pin) UNU;
-
+/*----------------------------------------------------------------------------*/
+// Function of pwm define
+/*----------------------------------------------------------------------------*/
+static int	pinToSysPwmPath	(int pin);
+static int	pwmSetup (int pin);
+static int	pwmRelease (int pin);
 /*----------------------------------------------------------------------------*/
 // wiringPi core function
 /*----------------------------------------------------------------------------*/
@@ -340,15 +341,109 @@ static int gpioToMuxReg (int pin)
 }
 
 /*----------------------------------------------------------------------------*/
-static int gpioToPwmPin (int pin)
+//
+// config pwm sys path. "/sys/class/pwm/pwmchip?"
+//
+/*----------------------------------------------------------------------------*/
+static int pinToSysPwmPath (int pin)
 {
-	return _gpioToPwmPin[pin - N2_GPIOA_PIN_START];
+	const char *pwmGroup;
+	char pwmLinkSrc[(BLOCK_SIZE / 8)];
+	char pwmPath[(BLOCK_SIZE / 8)];
+	int sz_link;
+
+	memset(pwmLinkSrc, 0, sizeof(pwmLinkSrc));
+	memset(pwmPath, 0, sizeof(pwmPath));
+
+	pwmGroup = pinToPwm[pin];
+	pwm = opendir("/sys/class/pwm");
+	if (pwm == NULL) {
+		printf("need to set device: pwm\n");
+		return -1;
+	}
+
+	while (1) {
+		pwmchip = readdir(pwm);
+
+		if (pwmchip == NULL) {
+			break;
+		}
+
+		if (strlen(pwmchip->d_name) <= 2)
+			continue;
+
+		sprintf(pwmPath, "%s/%s", "/sys/class/pwm", pwmchip->d_name);
+		sz_link = readlink(pwmPath, pwmLinkSrc, sizeof(pwmLinkSrc));
+		if (sz_link < 0) {
+			perror("Read symbolic link fail");
+			return sz_link;
+		}
+
+		if (strstr(pwmLinkSrc, pwmGroup) != NULL) {
+			strncpy(sysPwmPath, pwmPath, (sizeof(sysPwmPath) - 1));
+			break;
+		}
+	}
+	closedir(pwm);
+
+	return 0;
 }
 
-/*----------------------------------------------------------------------------*/
-static int gpioTophysPin (int pin)
-{
-	return _gpioTophysPin[pin - N2_GPIOA_PIN_START];
+static int pwmSetup (int pin) {
+	char cmd[(BLOCK_SIZE * 2)];
+	int pwmPin, ret;
+
+	memset(cmd, 0, sizeof(cmd));
+	memset(pwmExport, 0, sizeof(pwmExport));
+
+	if ((ret = pinToSysPwmPath(pin)) < 0) {
+		perror("set pwm dtb overlays");
+		return ret;
+	}
+
+	if (strstr(sysPwmPath, "pwmchip") == NULL) {
+		printf("config pwm dtb overlays\n");
+		return -1;
+	}
+
+	pwmPin = pinToPwmNum[pin];
+	pwmClock = N2_PWM_INTERNAL_CLK;
+	sprintf(pwmExport, "%d", (pwmPin % 2));
+	sprintf(pwmPinPath[pwmPin], "%s/pwm%d", sysPwmPath, (pwmPin % 2));
+	strncpy(setupedPwmPinPath[pwmPin], pwmPinPath[pwmPin], (BLOCK_SIZE - 1));
+#ifdef ANDROID
+	sprintf(cmd, "su -s sh -c %s %s", PWM_ACCESS_SCRIPT, pwmPinPath[pwmPin]);
+#else
+	sprintf(cmd, "sudo sh %s %s", PWM_ACCESS_SCRIPT, pwmPinPath[pwmPin]);
+#endif
+	inputToSysNode(sysPwmPath, "export", pwmExport);
+	system(cmd);
+	printf("PWM/pin%d: Don't change to gpio mode with overlay registered.\n", pin);
+
+	return 0;
+}
+
+static int pwmRelease (int pin) {
+	int pwmPin, ret;
+
+	if ((ret = pinToSysPwmPath(pin)) < 0) {
+		return ret;
+	}
+
+	if (strstr(sysPwmPath, "pwmchip") == NULL) {
+		return -1;
+	}
+
+	pwmPin = pinToPwmNum[pin];
+	sprintf(pwmUnexport, "%d", (pwmPin % 2));
+	sprintf(pwmPinPath[pwmPin], "%s/pwm%d", sysPwmPath, (pwmPin % 2));
+	if ((pwm = opendir(pwmPinPath[pwmPin])) != NULL) {
+		inputToSysNode(pwmPinPath[pwmPin], "enable", "0");
+		inputToSysNode(sysPwmPath, "unexport", pwmUnexport);
+		closedir(pwm);
+	}
+
+	return 0;
 }
 
 /*----------------------------------------------------------------------------*/
@@ -432,10 +527,10 @@ static int _pinMode (int pin, int mode)
 
 	if (lib->mode == MODE_GPIO_SYS)
 		return -1;
-
 	if ((pin = _getModeToGpio(lib->mode, pin)) < 0)
 		return -1;
 
+	pwmRelease (origPin);
 	softPwmStop  (origPin);
 	softToneStop (origPin);
 
@@ -471,29 +566,7 @@ static int _pinMode (int pin, int mode)
 		softToneCreate (origPin);
 		break;
 	case	PWM_OUTPUT:
-		usingGpiomemCheck("pinMode PWM");
-
-		int pwm_pin, alt;
-		pwm_pin = gpioToPwmPin(pin);
-		if( pwm_pin == -1 )
-		{
-			msg(MSG_WARN, "%s : This pin does not support hardware PWM mode.\n", __func__);
-			return -1;
-		}
-
-		alt		= pwmPinToALT[pwm_pin];
-		*(gpio + mux)	= (*(gpio + mux) & ~(0xF << target)) | (alt << target);
-
-#ifndef ANDROID
-		/**
-		 * 24 MHz / 120
-		 * 200 kHz / 500
-		 * frequency of PWM: 400 Hz
-		 * period of PWM: 2500 us
-		 */
-		_pwmSetClock(120);
-		_pwmSetRange(500);
-#endif
+		pwmSetup(origPin);
 		break;
 	default:
 		msg(MSG_WARN, "%s : Unknown Mode %d\n", __func__, mode);
@@ -639,29 +712,27 @@ static int _digitalWrite (int pin, int value)
 /*----------------------------------------------------------------------------*/
 static int _pwmWrite (int pin, int value)
 {
-	/**
-	 * @todo Add node
-	 * struct wiringPiNodeStruct *node = wiringPiNodes;
-	 */
+	unsigned int duty;
+	int pwmPin;
+
+	memset(pwmDuty, 0, sizeof(pwmDuty));
 
 	if (lib->mode == MODE_GPIO_SYS)
 		return -1;
 
-	if ((pin = _getModeToGpio(lib->mode, pin)) < 0)
+	if (((unsigned int)value > pwmRange) || (pwmRange <= 0)) {
+		printf("warn : pwm range value is greater than or equal pwmWrite's\n");
 		return -1;
-
-	int pwm_pin	= gpioToPwmPin(pin);
-	uint16_t range	= pwmPinToRange[pwm_pin];
-
-	if( value > range ) {
-		value = range;
 	}
 
-	*(pwm[pwm_pin/2] + pwmPinToDutyOffset[pwm_pin]) = (value << 16) | (range - value);
+	pwmPin = pinToPwmNum[pin];
+	duty = ((value * 100) / pwmRange);
+	sprintf(pwmDuty, "%d", ((atoi(pwmPeriod) * duty) / 100));
+
+	inputToSysNode(pwmPinPath[pwmPin], "duty_cycle", pwmDuty);
 
 	return 0;
 }
-
 /*----------------------------------------------------------------------------*/
 static int _analogRead (int pin)
 {
@@ -743,10 +814,35 @@ static int _digitalWriteByte (const unsigned int value)
 /*----------------------------------------------------------------------------*/
 static void _pwmSetRange (unsigned int range)
 {
-	range = range & 0xFFFF;
-	for( int i = 0; i < 6; ++i )
-	{
-		pwmPinToRange[i] = range;
+	unsigned int freq, period;
+
+	memset(pwmPeriod, 0, sizeof(pwmPeriod));
+
+	if (lib->mode == MODE_GPIO_SYS)
+		return;
+
+	if (pwmClock < 2) {
+		printf("error : pwm freq: %dMHz / (pwmSetClock's value) >= 2\n",
+				(N2_PWM_INTERNAL_CLK / 1000000));
+		return;
+	}
+
+	pwmRange = range;
+	if ((pwmRange < 1) || (pwmRange >= pwmClock)) {
+		printf("error : invalied value. ( < pwm freq)\n");
+		return;
+	}
+
+	freq = (pwmClock / pwmRange);
+	period = (1000000000 / freq); // period: s to ns.
+	sprintf(pwmPeriod, "%d", period);
+
+	for (int i = 0; i < 10; i++) {
+		if (strstr(setupedPwmPinPath[i], "None") != NULL)
+			continue;
+		inputToSysNode(setupedPwmPinPath[i], "period", pwmPeriod);
+		inputToSysNode(setupedPwmPinPath[i], "polarity", "normal");
+		inputToSysNode(setupedPwmPinPath[i], "enable", "1");
 	}
 }
 
@@ -757,27 +853,17 @@ static void _pwmSetRange (unsigned int range)
 /*----------------------------------------------------------------------------*/
 static void _pwmSetClock (int divisor)
 {
-	if((divisor < 1) || (divisor > 128))
-	{
-		msg(MSG_ERR,
-			"Set the clock prescaler (divisor) to 1 or more and 128 or less.: %s\n",
-			strerror (errno));
-	}
-	divisor = (divisor - 1);
-
-	for(uint16_t i = 1; i < 3; ++i) {
-		*( pwm[i] + N2_PWM_MISC_REG_01_OFFSET ) = \
-			(1 << N2_PWM_1_CLK_EN) \
-			| ( divisor << N2_PWM_1_CLK_DIV0) \
-			| (1 << N2_PWM_0_CLK_EN) \
-			| ( divisor << N2_PWM_0_CLK_DIV0) \
-			| (0 << N2_PWM_1_CLK_SEL0) \
-			| (0 << N2_PWM_0_CLK_SEL0) \
-			| (1 << N2_PWM_1_EN) \
-			| (1 << N2_PWM_0_EN);
+	if (pwmClock > 0)
+		pwmClock = (pwmClock / divisor);
+	else {
+		printf("error : pwm mode error\n");
+		return;
 	}
 }
 
+/*----------------------------------------------------------------------------*/
+
+
 /*----------------------------------------------------------------------------*/
 static unsigned int _digitalReadByte (void)
 {
@@ -826,12 +912,6 @@ static void init_gpio_mmap (void)
 			msg(MSG_ERR, "wiringPiSetup: mmap (GPIO) failed: %s \n", strerror (errno));
 		else
 			gpio = (uint32_t *) mapped;
-
-		for(uint16_t i = 1; i < 3; ++i) {
-			pwm[i] = ( uint32_t * )mmap( 0, BLOCK_SIZE, PROT_READ | PROT_WRITE, MAP_SHARED, fd, N2_GPIO_PWM_BASE + (0x1000 * (2 - i)) );
-			if( ( void * )pwm == MAP_FAILED )
-				msg(MSG_ERR, "wiringPiSetup: mmap (PWM) failed: %s \n", strerror (errno));
-		}
 	}
 }
 
