@@ -27,6 +27,7 @@
 #include <sys/wait.h>
 #include <sys/ioctl.h>
 #include <sys/utsname.h>
+#include <sys/types.h>
 #include <asm/ioctl.h>
 
 /*----------------------------------------------------------------------------*/
@@ -59,7 +60,7 @@ const char *piModelNames [16] =
 	"ODROID-C2",
 	"ODROID-XU3/XU4",
 	"ODROID-N1",
-	"ODROID-N2/N2Plus",
+	"ODROID-N2/N2Plus/N2L",
 	"ODROID-C4",
 	"ODROID-HC4",
 	"ODROID-M1",
@@ -223,7 +224,6 @@ int wiringPiFailure (int fatal, const char *message, ...)
 	va_end (argp) ;
 
 	fprintf (stderr, "%s", buffer);
-	exit (EXIT_FAILURE);
 
 	return 0 ;
 }
@@ -264,6 +264,30 @@ void setUsingGpiomem(const unsigned int value)
 	libwiring.usingGpiomem = value;
 }
 
+/*----------------------------------------------------------------------------*/
+/*
+ * input data to sys node.
+ */
+/*----------------------------------------------------------------------------*/
+int inputToSysNode (const char* sysPath, const char* node, char* data) {
+	char dest[(BLOCK_SIZE * 2)];
+	FILE *fd;
+
+	memset(dest, 0, sizeof(dest));
+
+	sprintf(dest, "%s/%s", sysPath, node);
+
+	if((fd = fopen(dest, "w")) == NULL) {
+		fprintf(stderr, "sys: Unable to open %s: %s\n",
+				dest, strerror (errno));
+		return -errno;
+	}
+	fprintf(fd, "%s\n", data);
+	fclose(fd);
+
+	return 0;
+}
+
 /*----------------------------------------------------------------------------*/
 /*
  * setKernelVersion:
@@ -1287,6 +1311,20 @@ int wiringPiSetupPhys (void)
  * Initialisation (again), however this time we are using the /sys/class/gpio
  *	interface to the GPIO systems - slightly slower, but always usable as
  *	a non-root user, assuming the devices are already exported and setup correctly.
+ *
+ *  -----------------------------------------------------------------------------
+ * Applies from "odroid-wiringpi" 3.14.1 and later.
+ *
+ * [The odroid hardware pwm uses sysnode, but it is separate from it.]
+ *	I decided to use the wiringpi service as gpiomem base for odroid,
+ *	but the pwm is using sysnode.
+ *
+ *	In addition, i designed the API in common so that devices such as gpio as well as pwm
+ *	can be directly accessed and used by sysnode.
+ *
+ *	"inputToSysNode" is what it is.
+ *
+ *	Authored by Steve Jeong <how2soft@gmail.com>, <steve.jeong@hardkernel.com>
  */
 /*----------------------------------------------------------------------------*/
 int wiringPiSetupSys (void)
